@@ -1,5 +1,5 @@
 /*
- * Ghost Port Scan is distributed under the GPL (GNU Public Licence)
+ * Ghost Port Scan is distributed under terms of the GPL (see COPYING)
  *
  * Homepage: gps.sourceforge.net
  * Contact.: whitehat@altern.org
@@ -24,25 +24,24 @@
 #include<netinet/ip.h>
 #include<netinet/tcp.h>
 // custom librairies
-#include "services.h"
 #include "gps.h"
+#include "services.h"
 #include "gateway.h"
+#include "timeout.h"
+#include "fwrd.h"
 
 
-int nbopt=0, nbpkt=0, scan_type=SYN_SCAN, rand_seed, serv_scan=1, nbports, sockets[2];
-int max_retries, max_timeout, ping_port=80, fragment=NO_FRAG, spoofing=MAC_SPOOF;
-u_int16_t *tab_response, fport=1, lport=1024;
-u_int32_t dst_ip=0, nb_src_ip=0, net_ip, netmask;
-struct pcap *descr;
-struct libnet_link_int *interface;
+int nbopt=0, nbports, nb_src_ip=0, rand_seed, sockets[2];
+int ping_port=80, fragment=NO_FRAG, spoofing=MAC_SPOOF, scan_type=SYN_SCAN;
+u_int32_t dst_ip=0;
 services_t *st;
-extern char *optarg;
+u_int16_t *tab_response;
 fwrd_elt *tab_fwrd;
 arp_elt *src_ip;
 port_elt *tab_port;
-struct in_addr tmp_inet;
 u_char dst_mac[ETH_ALEN];
 char errbuf[GPS_ERRBUF_SIZE], *eth=NULL;
+extern char *optarg;
 
 
 void main_sig(int sig)
@@ -61,19 +60,22 @@ int main(int argc, char **argv)
 	char str_src_ip[STR_IP_LEN], str_dst_ip[BUFSIZ];
 	char *ptr, *ptr1, *ptr2;
 	const u_char *packet;
-	u_int16_t port;
-	u_int32_t ip;
-	int npid, scan_speed=0, cmd_parse, i, j, k, gotcha[2], known, read_ok=1, i_ip=0, timeout=0, verbose=0;
+	u_int16_t port, fport=1, lport=1024;
+	u_int32_t ip, net_ip, netmask;
+	int npid, scan_speed=0, cmd_parse, i, j, gotcha[2], known, onlocalnet;
+	int i_ip=0, max_retries, max_timeout=0, verbose=0, nbpkt=0, serv_scan=1;
+	struct libnet_link_int *interface;
+	struct pcap *descr;
 	struct pcap_pkthdr phdr;
 	struct ether_header *ehdr;
 	struct ether_arp *ahdr, duplic;
 	struct ip *ihdr;
 	struct tcphdr *thdr;
-	struct timeval tv, current;
+	struct timeval tv;
 	sock_elt sckelt;
-	fwrd_elt tmp_fwrd;
 
 
+// ugly debugging code
 for(i=0; i<17; i++) signal(i, main_sig);
 
 
@@ -97,7 +99,7 @@ for(i=0; i<17; i++) signal(i, main_sig);
 				do
 				{
 					for(ptr=ptr1; *ptr && !isdigit(*ptr); ptr++);
-					if(!*ptr) error("Invalid list of src_ip", -1);
+					if(!*ptr) libnet_error(LIBNET_ERR_FATAL, "Invalid list of source IP addresses: %s\n", optarg);
 					for(ptr1=ptr2=ptr; *ptr1 && *ptr1!=',' && *ptr1!='-'; ptr1++)
 						if(*ptr1=='.') ptr2=ptr1;
 					switch(*ptr1)
@@ -106,7 +108,7 @@ for(i=0; i<17; i++) signal(i, main_sig);
 						case ',':
 						case 0:
 							if((i=ptr1-ptr)>=STR_IP_LEN)
-								error("Invalid list of src_ip", -1);
+								libnet_error(LIBNET_ERR_FATAL, "Invalid list of source IP addresses: %s\n", optarg);
 							strncpy(str_src_ip, ptr, i);
 							str_src_ip[i]='\0';
 							if(!(src_ip=(arp_elt *)realloc(src_ip, (nb_src_ip+1)*sizeof(arp_elt))))
@@ -140,9 +142,20 @@ for(i=0; i<17; i++) signal(i, main_sig);
 			case 'p':
 				// first_port and last_port
 				if(!(fport=atoi(optarg))) error("atoi: invalid first port", -1);
-				for(ptr=optarg; *ptr && *ptr!='-'; ptr++);
-				if(!(lport=atoi(++ptr))) error("atoi: invalid last port", -1);
-				if(lport<fport) error("Invalid first and last ports", -1);
+				for(ptr=optarg; isdigit(*ptr); ptr++);
+				switch(*ptr)
+				{
+					case 0:
+						lport=fport;
+						break;
+					case '-':
+						if(!(lport=atoi(++ptr))) error("atoi: invalid last port", -1);
+						if(lport<fport) error("Invalid first and last ports", -1);
+						break;
+					default:
+						libnet_error(LIBNET_ERR_FATAL, "Invalid argument: %s\n", optarg);
+						break;
+				}
 				serv_scan=0;
 				break;
 			case 'k':
@@ -250,16 +263,18 @@ for(i=0; i<17; i++) signal(i, main_sig);
 	fprintf(stdout, "Network device: %s\n", eth);
 
 
+	// get the local nework IP, and the netmask
+	if(pcap_lookupnet(eth, &net_ip, &netmask, errbuf))
+		libnet_error(LIBNET_ERR_FATAL, "pcap_lookupnet failed (%s)\n", errbuf);
+	// see if we can reach the host directly
+	onlocalnet=IS_ON_LOCAL_NET(dst_ip, netmask, net_ip);
+
 	// if MAC spoofing is activated, get the MAC address of
 	// the host or gateway to which send the ethernet trames
 	if(spoofing==MAC_SPOOF)
 	{
 		if(verbose) fprintf(stdout, "Searching for the destination MAC address\n");
-		// get the local nework IP, and the netmask
-		if(pcap_lookupnet(eth, &net_ip, &netmask, errbuf))
-			libnet_error(LIBNET_ERR_FATAL, "pcap_lookupnet failed (%s)\n", errbuf);
-		// see if we can reach the host directly
-		if(IS_ON_LOCAL_NET(dst_ip, netmask, net_ip))
+		if(onlocalnet)
 		{
 			if(get_mac(FARP, dst_ip, dst_mac, eth)==-1)
 				if(send_ARP_request(eth, &src_ip[0].ip, &dst_ip, src_ip[0].ea, dst_mac)==-1)
@@ -278,11 +293,31 @@ for(i=0; i<17; i++) signal(i, main_sig);
 		if(verbose>2)
 		{
 			fprintf(stdout, "Destination MAC address: ");
-			for(i=0; i<ETH_ALEN; i++) fprintf(stdout, "%2x%c", dst_mac[i], (i==ETH_ALEN-1)?'\n':':');
+			for(i=0; i<ETH_ALEN; i++) fprintf(stdout, "%x%c", dst_mac[i], (i==ETH_ALEN-1)?'\n':':');
 		}
 	}
 
 
+	// get a timeout value
+	if(verbose) fprintf(stdout, "Evaluating a timeout value..\n");
+	switch(spoofing)
+	{
+		case IP_SPOOF:
+			max_timeout=ip_eval_timeout(eth, rand_seed, verbose, src_ip[0].ip, dst_ip, ping_port, src_ip[0].ea);
+			break;
+		case MAC_SPOOF:
+			max_timeout=ethernet_eval_timeout(eth, rand_seed, verbose, src_ip[0].ip, dst_ip, ping_port, src_ip[0].ea, dst_mac);
+			break;
+	}
+	if(max_timeout==-1)
+	{
+		if(onlocalnet) max_timeout=DEFAULT_TIMEOUT_LOCAL;
+		else max_timeout=DEFAULT_TIMEOUT_DIST;
+		libnet_error(LIBNET_ERR_WARNING, "Timeout evaluation failure. Are we talking to a dead host ?\n");
+		fprintf(stderr, "Using the default timeout value: %d ms\n", max_timeout);
+	}
+
+
 	// clone the thread and setup a communication link between parent and child
 	// parent thread: sniffer
 	// child  thread: packet generator
@@ -290,6 +325,7 @@ for(i=0; i<17; i++) signal(i, main_sig);
 	if((npid=fork())==-1) error("fork: thread duplication failed", -1);
 	if(!npid) // child
 	{
+// ugly debugging code
 for(i=0; i<20; i++) signal(i, pg_sig);
 		close(sockets[0]);
 		switch(spoofing)
@@ -305,8 +341,15 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 	close(sockets[1]);
 
 	// allocate memory for data
-	if(!(tab_response=(u_int16_t *)malloc(nbports*sizeof(u_int16_t)))) error(NULL, -2);
-	if(scan_type==FWRD_SCAN) if(!(tab_fwrd=(fwrd_elt *)malloc(nb_src_ip*nbports*sizeof(fwrd_elt)))) error(NULL, -2);
+	if(scan_type!=FWRD_SCAN)
+	{
+		if(!(tab_response=(u_int16_t *)malloc(nbports*sizeof(u_int16_t)))) error(NULL, -2);
+	}
+	else
+	{
+		if(!(tab_fwrd=(fwrd_elt *)malloc(nb_src_ip*sizeof(fwrd_elt)))) error(NULL, -2);
+		for(i=0; i<nb_src_ip; i++) init_fwrd_elt(&tab_fwrd[i], src_ip[i].ip);
+	}
 
 	// open a raw link interface to send spoofed ARP replies if needed (something like ARP poisoning)
 	if(!(interface=libnet_open_link_interface(eth, errbuf)))
@@ -321,94 +364,11 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 	i=0; while(write(sockets[0], &sckelt, sizeof(sock_elt))<0 && i++<3);
 	if(i>2) error("write: operation on socket failed", -1);
 
-	// unblock the pcap socket
+	// unblock sockets
 	unblock_socket(descr->fd);
-
-	// evaluate timeout value
-	if(verbose>1) fprintf(stdout, "Evaluating a timeout value by TCP pinging port %d\n", ping_port);
-	i=k=0;
-	do
-	{
-		if(read_ok)
-		{
-			if(read(sockets[0], &sckelt, sizeof(sock_elt))>=0)
-			{
-				read_ok=0;
-				gettimeofday(&tv, NULL);
-			}
-			else continue;
-		}
-		if((packet=pcap_next(descr, &phdr)))
-		{
-			ehdr=(struct ether_header *)packet;
-			switch(ntohs(ehdr->ether_type))
-			{
-				case ETHERTYPE_ARP: // ARP packet
-					// send a fake ARP reply if needed
-					ahdr=(struct ether_arp *)((u_char *)ehdr + ETHER_HDR_LEN);
-					if(ntohs(ahdr->ea_hdr.ar_op)!=ARPOP_REQUEST) break;
-					if((j=search_src_ip(*((u_int32_t *)ahdr->arp_tpa)))==-1) break;
-					duplic=*ahdr;
-					ahdr->ea_hdr.ar_op=ntohs(ARPOP_REPLY);
-					for(k=0; k<ETH_ALEN; k++)
-					{
-						ahdr->arp_sha[k]=ehdr->ether_shost[k]=(u_int8_t)src_ip[j].ea[k];
-						ahdr->arp_tha[k]=ehdr->ether_dhost[k]=duplic.arp_sha[k];
-					}
-					for(k=0; k<4; k++)
-					{
-						ahdr->arp_spa[k]=duplic.arp_tpa[k];
-						ahdr->arp_tpa[k]=duplic.arp_spa[k];
-					}
-					if(libnet_write_link_layer(interface, eth, (u_char *)packet, ETHER_HDR_LEN + sizeof(struct ether_arp))==-1)
-						libnet_error(LIBNET_ERR_WARNING, "libnet_write_link layer failed\n");
-					break;
-				case ETHERTYPE_IP: // IP datagram
-					ihdr=(struct ip *)((u_char *)ehdr + ETHER_HDR_LEN);
-					if(ihdr->ip_src.s_addr!=dst_ip || search_src_ip(ihdr->ip_dst.s_addr)==-1 || ihdr->ip_p!=IPPROTO_TCP || (ihdr->ip_off & !IP_DF)) break;
-					thdr=(struct tcphdr *)((u_char *)ihdr + 4 * ihdr->ip_hl);
-					if(!(thdr->th_flags & TH_RST)) break;
-					gettimeofday(&current, NULL);
-					i++;
-					timeout+=TIMEVAL_MSEC_DIFF(current, sckelt.tv);
-					if(verbose>2) fprintf(stdout, "Time to get a RST back: %d ms\n", timeout);
-					if(i<3)
-					{
-						k=0; // reset the timeout counter
-						sckelt.i=1;
-						j=0; while(write(sockets[0], &sckelt, sizeof(sock_elt))<0 && j++<3);
-						if(j<3) read_ok=1;
-						else error("write: operation on socket failed", -1);
-					}
-					break;
-			}
-		}
-
-		// handle timeout
-		if(i<3)
-		{
-			gettimeofday(&current, NULL);
-			if(TIMEVAL_MSEC_DIFF(current, tv)>PING_TIMEOUT)
-			{
-				if(++k>3) error("Too much timeouts during the timeout evaluation", -1);
-				sckelt.i=1;
-				j=0; while(write(sockets[0], &sckelt, sizeof(sock_elt))<0 && j++<3);
-				if(j<3) read_ok=1;
-				else error("write: operation on socket failed", -1);
-			}
-		}
-	} while(i<3);
-	max_timeout=(TIMEOUT_FACTOR*timeout)/i;
-	// inform the packet generator that timeout evaluation is over
-	sckelt.i=i=0; while(write(sockets[0], &sckelt, sizeof(sock_elt))<0 && i++<3);
-	if(i>2) error("write: operation on socket failed", -1);
-
-	if(verbose>0) fprintf(stdout, "Timeout value: %d ms\n", max_timeout);
-
-
-	// unblock sockets[0]
 	unblock_socket(sockets[0]);
 
+	if(verbose) fprintf(stdout, "Starting port scan..\n");
 
 	// loop until all packets have been sent and all replies received
 	while(sckelt.i!=-2)
@@ -472,23 +432,25 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 						case FWRD_SCAN:			// FWRD scan
 							if(thdr->th_flags & TH_RST) i=1;
 							// we have a matching packet, see if we already have it
-							if(search_fwrd(ihdr->ip_dst.s_addr, thdr->th_sport)==-1)
+							port=ntohs(thdr->th_sport);
+							if(search_fwrd_port(&tab_fwrd[i_ip], port)==-1)
 							{
-								tab_fwrd[nbopt].ip.s_addr=ihdr->ip_dst.s_addr;
-								tab_fwrd[nbopt++].port=thdr->th_sport;
-								if((j=search_port(ntohs(thdr->th_sport)))!=-1) tab_port[j].gotcha=1;
+								nbopt++;
+								add_fwrd_port(&tab_fwrd[i_ip], port);
+								if((j=search_port(port))!=-1) tab_port[j].gotcha=1;
 							}
 							nbpkt++;
 							break;
 					}
 					if(i) break;
 					// we have a matching packet, see if we already have it
-					if((i=search_port(ntohs(thdr->th_sport)))!=-1)
+					port=ntohs(thdr->th_sport);
+					if((i=search_port(port))!=-1)
 					{
 						if(!tab_port[i].gotcha)
 						{
 							tab_port[i].gotcha=1;
-							tab_response[nbopt++]=thdr->th_sport;
+							tab_response[nbopt++]=port;
 						}
 					}
 					nbpkt++;
@@ -499,14 +461,11 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 		if(read(sockets[0], &sckelt, sizeof(sock_elt))>=0) switch(sckelt.i)
 		{
 			case -1: // first blast over
-				if(verbose>0)
+				if(verbose)
 				{
 					fprintf(stdout, "First blast over");
 					if(scan_type==FWRD_SCAN)
-					{
-						tmp_inet.s_addr=src_ip[i_ip].ip;
-						fprintf(stdout, " for %s", inet_ntoa(tmp_inet));
-					}
+						fprintf(stdout, " for %s", libnet_host_lookup(src_ip[i_ip].ip, LIBNET_RESOLVE));
 					fprintf(stdout, "\n");
 				}
 				break;
@@ -553,10 +512,7 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 				{
 					fprintf(stdout, "All packets");
 					if(scan_type==FWRD_SCAN)
-					{
-						tmp_inet.s_addr=src_ip[i_ip].ip;
-						fprintf(stdout, " for %s", inet_ntoa(tmp_inet));
-					}
+						fprintf(stdout, " for %s", libnet_host_lookup(src_ip[i_ip].ip, LIBNET_RESOLVE));
 					fprintf(stdout, " have been received\n");
 				}
 				gotcha[1]=1;
@@ -583,62 +539,33 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 	switch(scan_type)
 	{
 		case FWRD_SCAN: // fwrd scan
-			// order and handle byte swap
-			for(i=0; i<nbopt; i++)
-			{
-				tab_fwrd[i].port=ntohs(tab_fwrd[i].port);
-				tab_fwrd[i].ip.s_addr=ntohl(tab_fwrd[i].ip.s_addr);
-			}
-			for(i=0; i<nbopt; i++)
-				for(j=i+1; j<nbopt; j++)
-					if(tab_fwrd[i].ip.s_addr>tab_fwrd[j].ip.s_addr || (tab_fwrd[i].ip.s_addr==tab_fwrd[j].ip.s_addr && tab_fwrd[i].port>tab_fwrd[j].port))
-					{
-						tmp_fwrd=tab_fwrd[i];
-						tab_fwrd[i]=tab_fwrd[j];
-						tab_fwrd[j]=tmp_fwrd;
-					}
-			for(i=0; i<nbopt; i++) tab_fwrd[i].ip.s_addr=ntohl(tab_fwrd[i].ip.s_addr);
-			// output
 			if(nbopt==nb_src_ip*nbports)
 				fprintf(stdout, "None of the scanned ports are filtered for all specified IP addresses\n");
 			else
-				for(i=0; i<nbopt;)
+				for(i=0; i<nb_src_ip; i++)
 				{
-					for(j=i; j<nbopt && tab_fwrd[j].ip.s_addr==tab_fwrd[i].ip.s_addr; j++);
-					if((j-i)==nbports) fprintf(stdout, "None of the scanned ports are filtered for %s\n", inet_ntoa(tab_fwrd[i].ip));
-					else if(j-i<nbports/2)
-					{
-						fprintf(stdout, "Ports: ");
-						for(j=i; j<nbopt && tab_fwrd[j].ip.s_addr==tab_fwrd[i].ip.s_addr; j++)
-							fprintf(stdout, "%u, ", tab_fwrd[j].port);
-						fprintf(stdout, "\b\b are not filtered for %s\n", inet_ntoa(tab_fwrd[i].ip));
-					}
+					if(!tab_fwrd[i].nb_ports)
+						fprintf(stdout, "Each scanned port is filtered for %s\n", libnet_host_lookup(tab_fwrd[i].ip, LIBNET_RESOLVE));
+					else if(tab_fwrd[i].nb_ports==nbports)
+						fprintf(stdout, "Each scanned port is _not_ filtered for %s\n", libnet_host_lookup(tab_fwrd[i].ip, LIBNET_RESOLVE));
 					else
 					{
 						fprintf(stdout, "Ports: ");
-						if(st && serv_scan)
-							for(k=0; k<st->nb && st->tab[k].port<fport; k++)
-								if(search_fwrd(tab_fwrd[i].ip.s_addr, st->tab[i].port)==-1)
-									fprintf(stdout, "%u, ", st->tab[k].port);
-						for(port=fport; port<=lport; port++)
-							if(search_fwrd(tab_fwrd[k].ip.s_addr, port)==-1)
-								fprintf(stdout, "%u, ", port);
-						if(st && serv_scan)
+						if(tab_fwrd[i].nb_ports<=nbports/2)
 						{
-							for(k=0; k<st->nb && st->tab[k].port<=lport; k++);
-							for(; k<st->nb; k++)
-								if(search_fwrd(tab_fwrd[i].ip.s_addr, st->tab[k].port)==-1)
-									fprintf(stdout, "%u, ", st->tab[k].port);
+							fwrd_print_port_list(&tab_fwrd[i], "%d, ");
+							fprintf(stdout, "\b\b are _not_ filtered for %s\n", libnet_host_lookup(tab_fwrd[i].ip, LIBNET_RESOLVE));
+						}
+						else
+						{
+							fwrd_print_reverse_port_list(&tab_fwrd[i], "%d, ", tab_port, nbports);
+							fprintf(stdout, "\b\b are filtered for %s\n", libnet_host_lookup(tab_fwrd[i].ip, LIBNET_RESOLVE));
 						}
-						fprintf(stdout, "\b\b are filtered for %s\n", inet_ntoa(tab_fwrd[i].ip));
 					}
-					i=j;
 				}
 			break;
-
 		default:
-			// order and handle byte swap
-			for(i=0; i<nbopt; i++) tab_response[i]=ntohs(tab_response[i]);
+			// order
 			for(i=0; i<nbopt; i++)
 				for(j=i+1; j<nbopt; j++)
 					if(tab_response[i]>tab_response[j])
@@ -667,52 +594,38 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 				case FIN_SCAN:			// print results for FIN scan,
 				case XMAS_SCAN:			// for Xmas tree scan,
 				case NULL_SCAN:			// for Null scan,
-				case RAND_SCAN:		// for rand scan
+				case RAND_SCAN:			// for rand scan
 					if(nbopt==nbports) // no open ports found
 						fprintf(stdout, "No open ports found\n");
-					else
-					{
-						if(st && serv_scan)
-							for(i=0; i<st->nb && st->tab[i].port<fport; i++)
-								fin_scan_disp(st->tab[i].port);
-						for(port=fport; port<=lport; port++)
-							fin_scan_disp(port);
-						if(st && serv_scan)
-						{				
-							for(i=0; i<st->nb && st->tab[i].port<=lport; i++);
-							for(; i<st->nb; i++) fin_scan_disp(st->tab[i].port);
-						}
-					}
+					else if(!nbopt)
+						fprintf(stdout, "All scanned ports seem to be open\n");
+					else reverse_scan_disp();
 					break;
 				case TH_ACK: // ACK scan
 					if(!nbopt)
 						fprintf(stdout, "All scanned ports are filtered\n");
 					else if(nbopt==nbports)
 						fprintf(stdout, "None of the scanned ports are filtered\n");
-					else if(nbopt<nbports/2)
-					{
-						fprintf(stdout, "Ports: ");
-						for(i=0; i<nbopt; i++) fprintf(stdout, "%u, ", tab_response[i]);
-						fprintf(stdout, "\b\b are not filtered\n");
-					}
 					else
 					{
 						fprintf(stdout, "Ports: ");
-						if(st && serv_scan)
-							for(i=0; i<st->nb && st->tab[i].port<fport; i++)
-								if(search_tab(st->tab[i].port)==-1)
-									fprintf(stdout, "%u, ", st->tab[i].port);
-						for(port=fport; port<=lport; port++)
-							if(search_tab(port)==-1)
-								fprintf(stdout, "%u, ", port);
-						if(st && serv_scan)
+						if(nbopt<nbports/2)
 						{
-							for(i=0; i<st->nb && st->tab[i].port<=lport; i++);
-							for(; i<st->nb; i++)
-								if(search_tab(st->tab[i].port)==-1)
-									fprintf(stdout, "%u, ", st->tab[i].port);
+							for(i=0; i<nbopt; i++) fprintf(stdout, "%u, ", tab_response[i]);
+							fprintf(stdout, "\b\b are _not_ filtered\n");
+						}
+						else
+						{
+							for(i=j=0; i<nbports; i++)
+							{
+								if(tab_port[i].port<tab_response[j])
+									fprintf(stdout, "%d, ", tab_port[i].port);
+								else j++;
+							}
+							for(; i<nbports; i++)
+								fprintf(stdout, "%d, ", tab_port[i].port);
+							fprintf(stdout, "\b\b are filtered\n");
 						}
-						fprintf(stdout, "\b\b are filtered\n");
 					}
 					break;
 			}
@@ -721,8 +634,13 @@ for(i=0; i<20; i++) signal(i, pg_sig);
 	fprintf(stdout, "%d ports scanned\nGrabbed %d matching packets\n", nbports, nbpkt);
 
 	// free & close everything
-	if(scan_type==FWRD_SCAN) free(tab_fwrd);
-	free(tab_response);
+	if(scan_type==FWRD_SCAN)
+	{
+		for(i=0; i<nb_src_ip; i++)
+			free_fwrd_port_list(&tab_fwrd[i]);
+		free(tab_fwrd);
+	}
+	else free(tab_response);
 	free(tab_port);
 	free(src_ip);
 	pcap_close(descr);
@@ -764,36 +682,26 @@ int search_port(u_int16_t port)
 	return(-1);
 }
 
-int search_fwrd(u_int32_t ip, u_int16_t port)
+void reverse_scan_disp()
 {
-	int i;
-	for(i=0; i<nbopt && (tab_fwrd[i].ip.s_addr!=ip || tab_fwrd[i].port!=port); i++);
-	if(i<nbopt) return(i);
-	else return(-1);
-}
+	int i, j, known;
 
-void fin_scan_disp(u_int16_t pt)
-{
-	int known;
-	if(search_tab(pt)==-1)
+	for(i=j=0; i<nbports && j<nbopt; i++)
 	{
-		if((known=services_known(pt, st))==-1)
-			fprintf(stdout, "Port %u is open or filtered\n", pt);
-		else // port service detail
-			fprintf(stdout, "Port %u: %s (%s) is open or filtered\n", pt, st->tab[known].service, (st->tab[known].comment[0]) ? st->tab[known].comment : "-");
+		if(tab_port[i].port<tab_response[j])
+		{
+			if((known=services_known(tab_port[i].port, st))==-1)
+				fprintf(stdout, "Port %d is open or filtered\n", tab_port[i].port);
+			else fprintf(stdout, "Port %d: %s (%s) is open (or filtered)\n", tab_port[i].port, st->tab[known].service, (st->tab[known].comment[0]) ? st->tab[known].comment : "-");
+		}
+		else j++;
+	}
+	for(; i<nbports; i++)
+	{
+		if((known=services_known(tab_port[i].port, st))==-1)
+			fprintf(stdout, "%d\n", tab_port[i].port);
+		else fprintf(stdout, "Port %d: %s (%s) is open (or fileterd)\n", tab_port[i].port, st->tab[known].service, (st->tab[known].comment[0]) ? st->tab[known].comment : "-");
 	}
-}
-
-int search_tab(u_int16_t pt)
-{
-	int m, i=0, j=nbopt-1;
-	do {
-		m=(i+j)/2;
-		if(tab_response[m]==pt) return(m);
-		if(tab_response[m]<pt) i=m+1;
-		else j=m-1;
-	} while (i<=j);
-	return(-1);
 }
 
 int search_src_ip(u_int32_t ip)
@@ -810,18 +718,38 @@ void unblock_socket(int fd)
 	fcntl(fd, F_SETFL, options);
 }
 
+void randomize_port_order(u_int16_t *tab_switch_o2r, u_int16_t *tab_switch_r2o)
+{
+	int i, j, k;
+	port_elt tmp_port;
+
+	// randomize port order.
+	// Port order is randomized for the packet generator
+	// and ordered for the sniffer; tab_switch_r2o/o2r
+	// are used to translate randomized offset to ordered
+	// offset, and vice-versa
+	for(i=0; i<nbports; i++)
+	{
+		tmp_port=tab_port[i];
+		j=(nbports*libnet_get_prand(rand_seed))/255;
+		tab_port[i]=tab_port[j];
+		tab_port[j]=tmp_port;
+		k=tab_switch_r2o[i];
+		tab_switch_r2o[i]=tab_switch_r2o[j];
+		tab_switch_r2o[j]=k;
+	}
+	for(i=0; i<nbports; i++) tab_switch_o2r[tab_switch_r2o[i]]=i;
+}
 
 #define NB_RAND_FLAGS 8
 
 void ip_packet_generator(int scan_speed)
 {
-	int raw_sock, i, j, k, th_flags=scan_type, i_ip=0, offset;
+	int raw_sock, i, j, th_flags=scan_type, i_ip=0, offset;
 	int rflags[NB_RAND_FLAGS]={0, TH_FIN, TH_PUSH, TH_URG, TH_FIN|TH_PUSH, TH_FIN|TH_URG, TH_PUSH|TH_URG, TH_FIN|TH_PUSH|TH_URG};
 	char *packets=NULL, *frag_packets=NULL;
-	u_char *packet;
 	u_int32_t rsrc_ip=src_ip[0].ip;
 	sock_elt sckelt;
-	port_elt tmp_port;
 	u_int16_t *tab_switch_o2r=(u_int16_t *)malloc(nbports*sizeof(u_int16_t));
 	u_int16_t *tab_switch_r2o=(u_int16_t *)malloc(nbports*sizeof(u_int16_t));
 	if(!tab_switch_o2r||!tab_switch_r2o) error(NULL, -2);
@@ -837,45 +765,13 @@ void ip_packet_generator(int scan_speed)
 	// set up TCP flags for fwrd scan
 	if(scan_type==FWRD_SCAN) th_flags=TH_ACK;
 
-
-	// send some packets to evaluate a timeout value
-	libnet_init_packet(GPS_PACKET, &packet);
-	if(!packet) error(NULL, -2);
-	do
-	{
-		libnet_build_ip(GPS_TCP_H, IPTOS_LOWDELAY, libnet_get_prand(rand_seed)*libnet_get_prand(rand_seed), 0, 255, IPPROTO_TCP, src_ip[0].ip, dst_ip, NULL, 0, packet);
-		libnet_build_tcp(libnet_get_prand(rand_seed)+1024, ping_port, 65535*libnet_get_prand(rand_seed)*libnet_get_prand(rand_seed), 65535*libnet_get_prand(rand_seed)*libnet_get_prand(rand_seed), TH_ACK, 1024, 0, NULL, 0, packet+GPS_IP_H);
-		if(libnet_do_checksum(packet, IPPROTO_TCP, GPS_TCP_H)==-1)
-			libnet_error(LIBNET_ERR_FATAL, "libnet_do_checksum failed\n");
-		gettimeofday(&sckelt.tv, NULL);
-		i=0; while(libnet_write_ip(raw_sock, packet, GPS_PACKET)<GPS_PACKET && i++<3);
-		i=0; while(write(sockets[1], &sckelt, sizeof(sock_elt))<0 && i++<3);
-		i=0; while(read(sockets[1], &sckelt, sizeof(sock_elt))<0 && i++<3);
-	} while(sckelt.i);
-	libnet_destroy_packet(&packet);
-
 	// allocate memory for packets
 	if(!(packets=(char *)malloc(nbports*GPS_PACKET))) error(NULL, -2);
 	if(fragment!=NO_FRAG) if(!(frag_packets=(char *)malloc(nbports*GPS_PACKET))) error(NULL, -2);
 	do
 	{
 		// randomize port scan order.
-		// Port order is randomized for the packet generator
-		// and ordered for the sniffer; tab_switch_r2o/o2r
-		// are used to translate randomized offset to ordered
-		// offset, and vice-versa
-		for(i=0; i<nbports; i++)
-		{
-			tmp_port=tab_port[i];
-			j=(nbports*libnet_get_prand(rand_seed))/255;
-			tab_port[i]=tab_port[j];
-			tab_port[j]=tmp_port;
-			k=tab_switch_r2o[i];
-			tab_switch_r2o[i]=tab_switch_r2o[j];
-			tab_switch_r2o[j]=k;
-		}
-		for(i=0; i<nbports; i++) tab_switch_o2r[tab_switch_r2o[i]]=i;
-
+		randomize_port_order(tab_switch_o2r, tab_switch_r2o);
 
 		// build and send packets
 		offset=0;
@@ -984,14 +880,14 @@ void ip_packet_generator(int scan_speed)
 
 void ethernet_packet_generator(int scan_speed)
 {
-	int i, j, k, th_flags=scan_type, i_ip=0, offset;
+	int i, j, th_flags=scan_type, i_ip=0, offset;
 	int rflags[NB_RAND_FLAGS]={0, TH_FIN, TH_PUSH, TH_URG, TH_FIN|TH_PUSH, TH_FIN|TH_URG, TH_PUSH|TH_URG, TH_FIN|TH_PUSH|TH_URG};
 	char *packets=NULL, *frag_packets=NULL;
-	u_char *packet, *src_mac=src_ip[0].ea;
+	u_char *src_mac=src_ip[0].ea;
 	struct tcphdr *thdr;
+	struct libnet_link_int *interface;
 	u_int32_t rsrc_ip=src_ip[0].ip;
 	sock_elt sckelt;
-	port_elt tmp_port;
 	u_int16_t *tab_switch_o2r=(u_int16_t *)malloc(nbports*sizeof(u_int16_t));
 	u_int16_t *tab_switch_r2o=(u_int16_t *)malloc(nbports*sizeof(u_int16_t));
 	if(!tab_switch_o2r||!tab_switch_r2o) error(NULL, -2);
@@ -1007,46 +903,13 @@ void ethernet_packet_generator(int scan_speed)
 	// set up TCP flags for fwrd scan
 	if(scan_type==FWRD_SCAN) th_flags=TH_ACK;
 
-
-	// send some packets to evaluate a timeout value
-	libnet_init_packet(GPS_TRAME, &packet);
-	if(!packet) error(NULL, -2);
-	do
-	{
-		libnet_build_ip(GPS_TCP_H, IPTOS_LOWDELAY, libnet_get_prand(rand_seed)*libnet_get_prand(rand_seed), 0, 255, IPPROTO_TCP, src_ip[0].ip, dst_ip, NULL, 0, packet+GPS_ETH_H);
-		libnet_build_tcp(libnet_get_prand(rand_seed)+1024, ping_port, 65535*libnet_get_prand(rand_seed)*libnet_get_prand(rand_seed), 65535*libnet_get_prand(rand_seed)*libnet_get_prand(rand_seed), TH_ACK, 1024, 0, NULL, 0, packet+GPS_IP_H+GPS_ETH_H);
-		if(libnet_do_checksum(packet+GPS_ETH_H, IPPROTO_TCP, GPS_TCP_H)==-1 || libnet_do_checksum(packet+GPS_ETH_H, IPPROTO_IP, GPS_IP_H)==-1)
-			libnet_error(LIBNET_ERR_FATAL, "libnet_do_checksum failed\n");
-		libnet_build_ethernet(dst_mac, src_ip[0].ea, ETHERTYPE_IP, NULL, 0, packet);
-		gettimeofday(&sckelt.tv, NULL);
-		i=0; while(libnet_write_link_layer(interface, eth, packet, GPS_TRAME)<GPS_TRAME && i++<3);
-		i=0; while(write(sockets[1], &sckelt, sizeof(sock_elt))<0 && i++<3);
-		i=0; while(read(sockets[1], &sckelt, sizeof(sock_elt))<0 && i++<3);
-	} while(sckelt.i);
-	libnet_destroy_packet(&packet);
-
 	// allocate memory for packets
 	if(!(packets=(char *)malloc(nbports*GPS_TRAME))) error(NULL, -2);
 	if(fragment!=NO_FRAG) if(!(frag_packets=(char *)malloc(nbports*GPS_TRAME))) error(NULL, -2);
 	do
 	{
 		// randomize port scan order.
-		// Port order is randomized for the packet generator
-		// and ordered for the sniffer; tab_switch_r2o/o2r
-		// are used to translate randomized offset to ordered
-		// offset, and vice-versa
-		for(i=0; i<nbports; i++)
-		{
-			tmp_port=tab_port[i];
-			j=(nbports*libnet_get_prand(rand_seed))/255;
-			tab_port[i]=tab_port[j];
-			tab_port[j]=tmp_port;
-			k=tab_switch_r2o[i];
-			tab_switch_r2o[i]=tab_switch_r2o[j];
-			tab_switch_r2o[j]=k;
-		}
-		for(i=0; i<nbports; i++) tab_switch_o2r[tab_switch_r2o[i]]=i;
-
+		randomize_port_order(tab_switch_o2r, tab_switch_r2o);
 
 		// build and send packets
 		offset=0;
