@@ -10,7 +10,7 @@
 #ifdef VERSION
 #define GPS_VERSION VERSION
 #else
-#define GPS_VERSION "0.7.0"
+#define GPS_VERSION "0.7.1"
 #endif
 
 #define STR_IP_LEN 16
@@ -18,9 +18,6 @@
 #define TH_NULL 0
 
 
-#define PING_TIMEOUT	10000
-#define TIMEOUT_FACTOR	50
-
 // spoofing options
 #define MAC_SPOOF 1
 #define IP_SPOOF  0
@@ -47,12 +44,6 @@
 #define GPS_TCP_H (LIBNET_TCP_H)
 #define GPS_TRAME (GPS_ETH_H+GPS_IP_H+GPS_TCP_H)
 
-typedef struct
-{
-	struct in_addr ip;
-	u_int16_t port;
-} fwrd_elt;
-
 typedef struct
 {
 	u_int32_t ip;		// IP address
@@ -63,16 +54,16 @@ typedef struct
 
 typedef struct 
 {
-	u_int32_t ip; 	// src_ip GPS used
-	u_int16_t port;	// dst_port
-	u_int16_t src_port; // src_port
-	int seq;	// seq nb
-	int ack;	// ack nb
-	u_int16_t ip_id; // IP ID
-	int nbretries;	// nb of retries
-	struct timeval tv; // sending date
-	int gotcha;	// indicates if we have the packet
-	int wait;	// indicatates if the packet is queued
+	u_int32_t ip;		// src_ip GPS used
+	u_int16_t port;		// dst_port
+	u_int16_t src_port;	// src_port
+	int seq;		// seq nb
+	int ack;		// ack nb
+	int nbretries;		// nb of retries
+	struct timeval tv;	// sending date
+	int gotcha;		// indicates if we have the packet
+	int wait;		// indicatates if the packet is queued
+	u_int16_t ip_id;	// IP ID
 } port_elt;
 
 typedef struct
@@ -87,11 +78,10 @@ typedef struct
 void get_ea(char *ea);
 void init_port(int i, u_int16_t port);
 int  search_port(u_int16_t port);
-int  search_fwrd(u_int32_t ip, u_int16_t port);
-void fin_scan_disp(u_int16_t pt);
-int  search_tab(u_int16_t pt);
+void reverse_scan_disp();
 int  search_src_ip(u_int32_t ip);
 void unblock_socket(int fd);
+void randomize_port_order(u_int16_t *tab_switch_o2r, u_int16_t *tab_switch_r2o);
 void ip_packet_generator(int scan_speed);
 void ethernet_packet_generator(int scan_speed);
 void usage(char *cmd);
